import asyncio
import asyncclick as click
from bleak import (
    AdvertisementData,
    BleakClient,
    BleakError,
    BleakScanner,
    BLEDevice,
)
from bleak.uuids import uuidstr_to_str

from .const import FotaService, ScanService
from .parse import Characteristic, ManufacturerData, ProductType, Service


@click.group()
async def main():
    pass


@main.command()
async def scan():
    click.echo("Scanning for devices")

    devices = set()

    def detected(device: BLEDevice, advertisement: AdvertisementData):
        if device not in devices:
            if ScanService not in advertisement.service_uuids:
                return
            devices.add(device)

        click.echo(f"Device: {device}")
        for service in advertisement.service_uuids:
            click.echo(f" - Service: {service} {uuidstr_to_str(service)}")
        click.echo(f" - Data: {advertisement.service_data}")
        click.echo(f" - Manu: {advertisement.manufacturer_data}")

        if data := advertisement.manufacturer_data.get(ManufacturerData.company):
            decoded = ManufacturerData.decode(data)
            click.echo(f" -     : {decoded}")

        click.echo(f" - RSSI: {advertisement.rssi}")
        click.echo()

    async with BleakScanner(detected, service_uuids=[ScanService, FotaService]):
        while True:
            await asyncio.sleep(1)


@main.command()
@click.argument("address")
async def connect(address: str):
    click.echo(f"Connecting to: {address}")

    manufacturer_data = ManufacturerData()
    async with BleakScanner(service_uuids=[ScanService, FotaService]) as scanner:
        async with asyncio.timeout(10):
            async for device, data in scanner.advertisement_data():
                if device.address != address:
                    continue

                raw = data.manufacturer_data.get(ManufacturerData.company)
                if raw is None:
                    continue

                manufacturer_data.update(raw)

                product_type = ProductType.from_manufacturer_data(manufacturer_data)
                if product_type is ProductType.UNKNOWN:
                    continue

                break

    click.echo(f"Advertised data: {manufacturer_data}")
    click.echo(f"Product type: {product_type}")

    async with BleakClient(device, timeout=20) as client:
        for service in client.services:
            service_parser = Service.find_service(service.uuid, product_type)

            click.echo(
                f"Service: {service.uuid}: {service_parser.__name__ if service_parser else service.description}"
            )

            for char in service.characteristics:
                char_parser = None
                if service_parser:
                    char_parser = service_parser.characteristics.get(char.uuid)

                click.echo(
                    f" -  {char.uuid}: {char_parser.name if char_parser else char.description}"
                )
                click.echo(f"    Prop: {char.properties}")

                data = None
                if "read" in char.properties:
                    try:
                        data = await client.read_gatt_char(char.uuid)
                    except BleakError as exc:
                        click.echo(f"    Failed: {repr(exc)}")

                if data is not None:
                    click.echo(f"    Raw: {data}")
                    if char_parser:
                        click.echo(f"    Data: {char_parser.decode(data)!r}")


@main.command()
async def chars():
    for char in Characteristic.registry.values():
        click.echo(char.name)


try:
    main()
except KeyboardInterrupt:
    pass



from abc import ABC
from dataclasses import dataclass, field
from datetime import datetime, timezone
from enum import IntEnum, Enum, auto
from typing import ClassVar, Generic, Self, TypeVar


def pretty_name(name: str):
    data = name.split("_")
    return " ".join(f"{part[0].upper()}{part[1:]}" for part in data)


class ProductType(Enum):
    UNKNOWN = auto()
    MOWER = auto()
    WATER_COMPUTER = auto()
    VALVE = auto()
    PUMP = auto()
    PRESSURE_TANKS = auto()
    AQUA_CONTOURS = auto()
    AUTOMATS = auto()

    @staticmethod
    def from_manufacturer_data(data: "ManufacturerData") -> "ProductType":
        if data.group == 10:
            return ProductType.MOWER

        if data.group == 18:
            if data.model in (0, 1) and data.variant == 1:
                return ProductType.WATER_COMPUTER
            if data.model == 2 and data.variant == 1:
                return ProductType.VALVE
            if data.model == 16:
                return ProductType.AQUA_CONTOURS
            return ProductType.UNKNOWN

        if data.group == 17:
            if data.model == 1:
                return ProductType.PUMP
            if data.model == 2:
                return ProductType.PRESSURE_TANKS
            if data.model == 3:
                return ProductType.AUTOMATS
            return ProductType.UNKNOWN

        return ProductType.UNKNOWN


CharacteristicType = TypeVar("CharacteristicType")


@dataclass
class Characteristic(Generic[CharacteristicType]):
    uuid: str
    variant: str | None = None
    name: str = ""
    registry: ClassVar[dict[str, list[Self]]] = {}
    unique_id: str = field(init=False)

    def __set_name__(self, _, name: str):
        self.name = pretty_name(name)

    def __post_init__(self):
        if self.variant is not None:
            unique_id = self.uuid + ":" + self.variant
        else:
            unique_id = self.uuid
        object.__setattr__(self, "unique_id", unique_id)
        self.registry.setdefault(self.uuid, []).append(self)

    @classmethod
    def decode(cls, data: bytes) -> CharacteristicType:
        raise NotImplementedError(f"Decoding of {type(cls)} is not implemented")

    @classmethod
    def encode(cls, data: CharacteristicType) -> bytes:
        raise NotImplementedError(f"Encoding of {type(cls)} is not implemented")


@dataclass
class CharacteristicBytes(Characteristic[bytes]):
    @classmethod
    def decode(cls, data: bytes) -> bytes:
        return data

    @classmethod
    def encode(cls, value: bytes) -> bytes:
        return value


@dataclass
class CharacteristicBool(Characteristic[bool]):
    @classmethod
    def decode(cls, data: bytes) -> bool:
        return data[0] != 0

    @classmethod
    def encode(cls, data: bool) -> bytes:
        if data:
            return b"\x01"
        return b"\x00"


@dataclass
class CharacteristicString(Characteristic[str]):
    @classmethod
    def decode(cls, data: bytes) -> str:
        return data.decode("ASCII", "replace")

    @classmethod
    def encode(cls, value: str) -> bytes:
        return value.encode("ASCII")


@dataclass
class CharacteristicNullString(Characteristic[str]):
    @classmethod
    def decode(cls, data: bytes) -> str:
        return data.partition(b"\x00")[0].decode("ASCII", "replace")

    @classmethod
    def encode(cls, value: str) -> bytes:
        return value.encode("ASCII")


@dataclass
class CharacteristicNullStringUf8(Characteristic[str]):
    @classmethod
    def decode(cls, data: bytes) -> str:
        return data.partition(b"\x00")[0].decode("utf-8", "replace")

    @classmethod
    def encode(cls, value: str) -> bytes:
        return value.encode("utf-8")


@dataclass
class CharacteristicInt(Characteristic[int]):
    @classmethod
    def decode(cls, data: bytes) -> int:
        return int.from_bytes(data, "little", signed=True)

    @classmethod
    def encode(cls, value: int) -> bytes:
        return value.to_bytes(1, "little", signed=True)


@dataclass
class CharacteristicLong(Characteristic[int]):
    @classmethod
    def decode(cls, data: bytes) -> int:
        return int.from_bytes(data, "little", signed=True)

    @classmethod
    def encode(cls, value: int) -> bytes:
        return value.to_bytes(4, "little", signed=True)


@dataclass
class CharacteristicUInt16(Characteristic[int]):
    @classmethod
    def decode(cls, data: bytes) -> int:
        return int.from_bytes(data, "little", signed=False)

    @classmethod
    def encode(cls, value: int) -> bytes:
        return value.to_bytes(2, "little", signed=False)


@dataclass
class CharacteristicLongArray(Characteristic[list[int]]):
    @classmethod
    def decode(cls, data: bytes) -> list[int]:
        return [
            int.from_bytes(data[i : i + 4], "little", signed=True)
            for i in range(0, len(data), 4)
        ]


@dataclass
class CharacteristicWeekday(Characteristic[list[bool]]):
    @classmethod
    def decode(cls, data: bytes) -> list[bool]:
        value = int.from_bytes(data, "little", signed=False)
        return [(value >> i) & 1 != 0 for i in range(7)]

    @classmethod
    def encode(cls, value: list[bool]) -> bytes:
        int_value = 0
        for i, day in enumerate(value):
            if day:
                int_value |= 1 << i
        return int_value.to_bytes(1, "little", signed=False)


@dataclass
class CharacteristicTime(Characteristic[datetime]):
    @classmethod
    def decode(cls, data: bytes) -> datetime:
        value = int.from_bytes(data, "little")
        return datetime.fromtimestamp(value, timezone.utc).replace(tzinfo=None)

    @classmethod
    def encode(cls, value: datetime) -> bytes:
        return int(value.replace(tzinfo=timezone.utc).timestamp()).to_bytes(
            4, "little", signed=True
        )


@dataclass
class CharacteristicTimeArray(Characteristic[list[datetime]]):
    @classmethod
    def decode(cls, data: bytes) -> list[datetime]:
        return [
            datetime.fromtimestamp(value, timezone.utc).replace(tzinfo=None)
            for value in CharacteristicLongArray.decode(data)
        ]


@dataclass
class CharacteristicIntEnum[T: IntEnum](CharacteristicInt):
    enum: type[T] = field(kw_only=True)

    def decode(self, data: bytes) -> T | int:
        raw = int.from_bytes(data, "little", signed=True)
        try:
            return self.enum(raw)
        except ValueError:
            return raw

    def encode(self, value: T | int) -> bytes:
        return value.to_bytes(1, "little", signed=True)


@dataclass
class ErrorData[T: IntEnum]:
    current_event_index: int
    total_events: int
    time_stamp: datetime
    error_code: T | int


@dataclass
class CharacteristicErrorData[T: IntEnum](Characteristic[ErrorData[T]]):
    enum: type[T] = field(kw_only=True)

    def decode(self, data: bytes) -> ErrorData[T]:
        error_code = int.from_bytes(data[6:7], "little")
        try:
            error_code = self.enum(error_code)
        except ValueError:
            pass

        return ErrorData(
            int.from_bytes(data[0:1], "little"),
            int.from_bytes(data[1:2], "little"),
            datetime.fromtimestamp(
                int.from_bytes(data[2:6], "little"), timezone.utc
            ).replace(tzinfo=None),
            error_code,
        )

    @classmethod
    def encode(cls, value: ErrorData[T]) -> bytes:
        return [
            *value.current_event_index.to_bytes(1, "little", signed=True),
            *value.total_events.to_bytes(1, "little", signed=True),
            *int(value.time_stamp.replace(tzinfo=timezone.utc).timestamp()).to_bytes(
                4, "little", signed=True
            ),
            *value.error_code.to_bytes(1, "little", signed=True),
        ]


class Service:
    unique_id: ClassVar[str]
    uuid: ClassVar[str]
    variant: ClassVar[str | None] = None
    products: ClassVar[set[ProductType]] = set(ProductType)
    registry: ClassVar[dict[str, list[Self]]] = {}
    characteristics: ClassVar[dict[str, Characteristic]] = {}

    @classmethod
    def find_service(cls, uuid: str, product_type: ProductType) -> Self | None:
        services = cls.registry.get(uuid, [])
        for service in services:
            if product_type in service.products:
                return service
        return None

    @classmethod
    def services_for_product_type(cls, product_type: ProductType) -> list[Self]:
        """Get all services for a product type."""
        return [
            service
            for services in Service.registry.values()
            for service in services
            if product_type in service.products
        ]

    def __init_subclass__(cls, /, **kwargs):
        super().__init_subclass__(**kwargs)
        if ABC in cls.__bases__:
            return
        if cls.variant is not None:
            cls.unique_id = cls.uuid + ":" + cls.variant
        else:
            cls.unique_id = cls.uuid
        cls.registry.setdefault(cls.uuid, []).append(cls)

        cls.characteristics = {}
        for value in vars(cls).values():
            if isinstance(value, Characteristic):
                cls.characteristics[value.uuid] = value


class EnumOrInt(IntEnum):
    @classmethod
    def enum_or_int(cls, value: int):
        try:
            return cls(value)
        except ValueError:
            return value


class ProductGroup(EnumOrInt):
    MOWER = 10
    GARDEN_PUMP = 17
    WATER_CONTROL = 18


@dataclass
class ManufacturerData:
    company: ClassVar[int] = 0x0426
    pairable: bool | None = None
    serial: int | None = None
    group: int | ProductGroup | None = None
    model: int | None = None
    variant: int | None = None

    @staticmethod
    def decode_dict(data: bytes):
        res: dict[int, bytes] = {}
        idx = 0
        while idx < len(data):
            size = data[idx]
            key = data[idx + 1]
            res[key] = data[idx + 2 : idx + size + 1]
            idx += size + 1
        return res

    @staticmethod
    def decode(data: bytes):
        res = ManufacturerData()
        res.update(data)
        return res

    def update(self, data: bytes):
        value = ManufacturerData.decode_dict(data)
        info = dict(enumerate(value.get(6, b"")))

        if (data := info.get(0)) is not None:
            self.group = ProductGroup.enum_or_int(data)
        if (data := info.get(1)) is not None:
            self.model = data
        if (data := info.get(2)) is not None:
            self.variant = data

        if (data := value.get(4)) is not None:
            self.serial = int.from_bytes(data, "little")
        if (data := value.get(5)) is not None:
            self.pairable = bool.from_bytes(data, "little")



import asyncio
import logging
from collections.abc import Awaitable, Callable
from contextlib import asynccontextmanager
from datetime import datetime
from typing import TypeVar, overload

from bleak import BleakClient
from bleak.exc import BleakError
from bleak.backends.device import BLEDevice
from bleak_retry_connector import establish_connection

from .exceptions import (
    CharacteristicNoAccess,
    CharacteristicNotFound,
    CommunicationFailure,
)
from .parse import (
    Characteristic,
    CharacteristicTime,
    CharacteristicType,
    ProductType,
    Service,
)

LOGGER = logging.getLogger(__name__)
DEFAULT_MISSING = object()
DEFAULT_TYPE = TypeVar("DEFAULT_TYPE")
DEFAULT_DELAY = 1


class CallLaterJob:
    """Helper to contain a function that is to be called later."""

    def __init__(self, fun: Callable[[], Awaitable[None]]) -> None:
        """Initialize a call later job."""
        self._fun = fun
        self._cancel: asyncio.TimerHandle | asyncio.Task | None = None

    def cancel(self):
        """Cancel any pending delay call."""
        if self._cancel:
            self._cancel.cancel()
            self._cancel = None

    async def _call(self):
        await self.call_now()

    async def call_now(self):
        """Call function now."""
        self.cancel()
        await self._fun()

    def call_later(self, delay: float):
        """Call function sometime later."""
        self.cancel()

        async def _call_task():
            await self._fun()
            self._cancel = None

        def _call():
            self._cancel = asyncio.create_task(_call_task())

        self._cancel = asyncio.get_event_loop().call_later(delay, _call)


class CachedConnection:
    """Recursive and delay closed client."""

    def __init__(
        self, disconnect_delay: float, device_lookup: Callable[[], BLEDevice]
    ) -> None:
        """Initialize cached client."""

        self._client: BleakClient | None = None
        self._lock = asyncio.Lock()
        self._count = 0
        self._lookup = device_lookup
        self._disconnect_delay = disconnect_delay
        self._disconnect_job = CallLaterJob(self._disconnect)

    async def disconnect(self):
        await self._disconnect_job.call_now()

    async def _disconnect(self):
        async with self._lock:
            if client := self._client:
                LOGGER.debug("Disconnecting from %s", self._client.address)
                self._client = None
                await client.disconnect()

    async def _connect(self) -> BleakClient:
        device = self._lookup()

        LOGGER.debug("Connecting to %s", device.address)
        self._client = await establish_connection(
            BleakClient, device, "Gardena Bluetooth", use_services_cache=True
        )
        LOGGER.debug("Connected to %s", device.address)
        return self._client

    @asynccontextmanager
    async def __call__(self):
        """Retrieve a context manager for a cached client."""
        self._disconnect_job.cancel()

        try:
            async with self._lock:
                if not (client := self._client) or not client.is_connected:
                    client = await self._connect()

                self._count += 1
                try:
                    yield client
                finally:
                    self._count -= 1

                    if not self._count and self._client:
                        self._disconnect_job.call_later(self._disconnect_delay)
        except BleakError as exception:
            await self._disconnect_job.call_now()
            LOGGER.debug("Unexpected disconnection from device %s", exception)
            raise CommunicationFailure(
                f"Communcation failed with device: {exception}"
            ) from exception


class Client:
    def __init__(
        self,
        client_or_device: CachedConnection | BLEDevice,
        product_type: ProductType = ProductType.UNKNOWN,
    ) -> None:
        if isinstance(client_or_device, CachedConnection):
            self._client = client_or_device
        else:
            self._client = CachedConnection(DEFAULT_DELAY, lambda: client_or_device)

        self._services = Service.services_for_product_type(product_type)
        self._product_type = product_type
        self._unique_id = {
            char.unique_id
            for service in self._services
            for char in service.characteristics.values()
        }

    async def disconnect(self):
        await self._client.disconnect()

    @overload
    async def read_char_raw(self, uuid: str) -> bytes: ...

    @overload
    async def read_char_raw(
        self, uuid: str, default: DEFAULT_TYPE
    ) -> bytes | DEFAULT_TYPE: ...

    async def read_char_raw(
        self, uuid: str, default: DEFAULT_TYPE = DEFAULT_MISSING
    ) -> bytes | DEFAULT_TYPE:
        async with self._client() as client:
            characteristic = client.services.get_characteristic(uuid)
            if characteristic is None:
                if default is not DEFAULT_MISSING:
                    return default
                raise CharacteristicNotFound(f"Unable to find characteristic {uuid}")
            if "read" not in characteristic.properties:
                if default is not DEFAULT_MISSING:
                    return default
                raise CharacteristicNoAccess(f"Characteristic {uuid} is not readable")
            return await client.read_gatt_char(characteristic)

    @overload
    async def read_char(
        self, char: Characteristic[CharacteristicType]
    ) -> CharacteristicType: ...

    @overload
    async def read_char(
        self,
        char: Characteristic[CharacteristicType],
        default: DEFAULT_TYPE,
    ) -> CharacteristicType | DEFAULT_TYPE: ...

    async def read_char(
        self,
        char: Characteristic[CharacteristicType],
        default: DEFAULT_TYPE = DEFAULT_MISSING,
    ) -> CharacteristicType | DEFAULT_TYPE:
        """Read data to from a characteristic."""
        if char.unique_id not in self._unique_id:
            LOGGER.debug("Attempt to read unsupported %s", char.unique_id)
            if default is not DEFAULT_MISSING:
                return default
            raise CharacteristicNotFound

        try:
            return char.decode(await self.read_char_raw(char.uuid))
        except CharacteristicNotFound:
            if default is not DEFAULT_MISSING:
                return default
            raise

    async def write_char_raw(
        self, uuid: str, data: bytes, response: bool | None = None
    ):
        async with self._client() as client:
            """Write data to a characteristic."""
            characteristic = client.services.get_characteristic(uuid)
            if characteristic is None:
                raise CharacteristicNotFound(f"Unable to find characteristic {uuid}")

            if response is None:
                response = "write" in characteristic.properties

            if response:
                if "write" not in characteristic.properties:
                    raise CharacteristicNoAccess(
                        f"Characteristic {uuid} is not writable"
                    )
            else:
                if "write-without-response" not in characteristic.properties:
                    raise CharacteristicNoAccess(
                        f"Characteristic {uuid} is not writable without response"
                    )

            await client.write_gatt_char(characteristic, data, response=response)

    async def write_char(
        self,
        char: Characteristic[CharacteristicType],
        value: CharacteristicType,
        response: bool | None = None,
    ) -> None:
        """Write data to a characteristic."""
        if char.unique_id not in self._unique_id:
            raise CharacteristicNotFound

        data = char.encode(value)
        await self.write_char_raw(char.uuid, data, response)

    async def update_timestamp(self, char: CharacteristicTime, now: datetime):
        try:
            timestamp = await self.read_char(char)
        except CharacteristicNoAccess:
            LOGGER.debug("No timestamp defined for device")
            return
        timestamp = timestamp.replace(tzinfo=now.tzinfo)
        delta = timestamp - now
        if abs(delta.total_seconds()) > 60:
            LOGGER.warning(
                "Updating time on device to match local time delta was %s", delta
            )
            await self.write_char(
                char,
                now.replace(tzinfo=None),
                True,
            )
        else:
            LOGGER.debug("No need to update timestamp local time delta was %s", delta)

    async def get_all_characteristics(self) -> dict[str, Characteristic]:
        """Get all characteristics from device."""
        uuids = await self.get_all_characteristics_uuid()
        characteristics = {
            char.unique_id: char
            for service in self._services
            for char in service.characteristics.values()
            if char.uuid in uuids
        }
        return characteristics

    async def get_all_characteristics_uuid(self) -> set[str]:
        """Get all characteristics from device."""
        async with self._client() as client:
            characteristics = {
                characteristic.uuid
                for service in client.services
                for characteristic in service.characteristics
            }
            LOGGER.debug("Characteristics: %s", characteristics)
            return characteristics



from gardena_bluetooth.const import (
    Schedule_1,
    Schedule,
    Schedule_2,
    Schedule_3,
    Schedule_4,
    Schedule_5,
)
from gardena_bluetooth.parse import Service, Characteristic
import pytest


@pytest.mark.parametrize(
    "schedule,base",
    [
        (Schedule_1, "1"),
        (Schedule_2, "2"),
        (Schedule_3, "3"),
        (Schedule_4, "4"),
        (Schedule_5, "5"),
    ],
)
def test_schedule(schedule: type[Schedule], base: str):
    assert schedule.uuid == f"98bd0c{base}0-0b0e-421a-84e5-ddbf75dc6de4"
    assert schedule.start_time.uuid == f"98bd0c{base}1-0b0e-421a-84e5-ddbf75dc6de4"
    assert schedule.duration.uuid == f"98bd0c{base}2-0b0e-421a-84e5-ddbf75dc6de4"
    assert schedule.weekdays.uuid == f"98bd0c{base}3-0b0e-421a-84e5-ddbf75dc6de4"
    assert schedule.valve_link.uuid == f"98bd0c{base}4-0b0e-421a-84e5-ddbf75dc6de4"
    assert schedule.active.uuid == f"98bd0c{base}5-0b0e-421a-84e5-ddbf75dc6de4"
    assert schedule.sensor_link.uuid == f"98bd0c{base}6-0b0e-421a-84e5-ddbf75dc6de4"


def test_id_uniqueness():
    """Ensure our id's are globally unique for services and characteristics."""
    ids: dict[str, Service | Characteristic] = {}
    for services in Service.registry.values():
        for service in services:
            assert ids.setdefault(service.unique_id, service) is service

            for char in service.characteristics.values():
                assert ids.setdefault(char.unique_id, char) is char



class GardenaBluetoothException(Exception):
    pass


class CharacteristicNoAccess(GardenaBluetoothException):
    pass


class CharacteristicNotFound(CharacteristicNoAccess):
    pass


class CommunicationFailure(GardenaBluetoothException):
    pass



from abc import ABC
from enum import IntEnum
from .parse import (
    CharacteristicBool,
    CharacteristicWeekday,
    CharacteristicBytes,
    CharacteristicErrorData,
    CharacteristicInt,
    CharacteristicIntEnum,
    CharacteristicLong,
    CharacteristicLongArray,
    CharacteristicString,
    CharacteristicTime,
    CharacteristicNullString,
    CharacteristicNullStringUf8,
    CharacteristicTimeArray,
    CharacteristicUInt16,
    Service,
    ProductType,
)

PRODUCT_NAMES = {
    ProductType.PUMP: "Gardena Garden Pump",
    ProductType.WATER_COMPUTER: "Gardena Water Computer",
    ProductType.VALVE: "Gardena Irrigation Valve",
    ProductType.MOWER: "Gardena Mower",
    ProductType.AQUA_CONTOURS: "Gardena Aqua Precise",
    ProductType.AUTOMATS: "Gardena Automats",
    ProductType.PRESSURE_TANKS: "Gardena Pressure Tanks",
    ProductType.UNKNOWN: "Gardena Unknown Product",
}

ScanService = "98bd0001-0b0e-421a-84e5-ddbf75dc6de4"
FotaService = "0000ffc0-0000-1000-8000-00805f9b34fb"


class Scan(Service):
    uuid = "98bd0001-0b0e-421a-84e5-ddbf75dc6de4"

    write_characteristic = CharacteristicBytes("98bd0002-0b0e-421a-84e5-ddbf75dc6de4")
    read_characteristic = CharacteristicBytes("98bd0003-0b0e-421a-84e5-ddbf75dc6de4")
    read_protocol_descriptor = CharacteristicNullString(
        "98bd0004-0b0e-421a-84e5-ddbf75dc6de4"
    )


class Valve(Service):
    uuid = "98bd0f10-0b0e-421a-84e5-ddbf75dc6de4"

    state = CharacteristicBool("98bd0f11-0b0e-421a-84e5-ddbf75dc6de4")
    connected_state = CharacteristicBool("98bd0f12-0b0e-421a-84e5-ddbf75dc6de4")
    remaining_open_time = CharacteristicLong("98bd0f13-0b0e-421a-84e5-ddbf75dc6de4")
    manual_watering_time = CharacteristicLong("98bd0f14-0b0e-421a-84e5-ddbf75dc6de4")
    activation_reason = CharacteristicInt("98bd0f15-0b0e-421a-84e5-ddbf75dc6de4")


class DeviceConfiguration(Service):
    uuid = "98bd0b10-0b0e-421a-84e5-ddbf75dc6de4"
    products = set(ProductType) - {ProductType.AQUA_CONTOURS}

    rain_pause = CharacteristicLong("98bd0b11-0b0e-421a-84e5-ddbf75dc6de4")
    seasonal_adjust = CharacteristicInt("98bd0b12-0b0e-421a-84e5-ddbf75dc6de4")
    unix_timestamp = CharacteristicTime("98bd0b13-0b0e-421a-84e5-ddbf75dc6de4")
    mobile_device_name = CharacteristicInt("98bd0b14-0b0e-421a-84e5-ddbf75dc6de4")
    device_language = CharacteristicInt("98bd0b15-0b0e-421a-84e5-ddbf75dc6de4")
    display_brightness = CharacteristicInt("98bd0b16-0b0e-421a-84e5-ddbf75dc6de4")
    first_user_start = CharacteristicBool("98bd0b17-0b0e-421a-84e5-ddbf75dc6de4")
    custom_device_name = CharacteristicNullStringUf8(
        "98bd0b18-0b0e-421a-84e5-ddbf75dc6de4"
    )


class AquaContourContours(Service):
    uuid = "98bd0b10-0b0e-421a-84e5-ddbf75dc6de4"
    products = {ProductType.AQUA_CONTOURS}
    variant = "1"

    contour_receive = CharacteristicBytes(
        "98bd0b11-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    contour_transmit = CharacteristicBytes(
        "98bd0b12-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    contour_info = CharacteristicBytes(
        "98bd0b13-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    contour_name_1 = CharacteristicNullStringUf8(
        "98bd0b1a-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    contour_name_2 = CharacteristicNullStringUf8(
        "98bd0b1b-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    contour_name_3 = CharacteristicNullStringUf8(
        "98bd0b1c-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    contour_name_4 = CharacteristicNullStringUf8(
        "98bd0b1d-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    contour_name_5 = CharacteristicNullStringUf8(
        "98bd0b1e-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )


class AquaContourSchedule(Service):
    uuid = "98bd0c10-0b0e-421a-84e5-ddbf75dc6de4"
    products = {ProductType.AQUA_CONTOURS}
    variant = "1"

    schedule_1 = CharacteristicBytes(
        "98bd0c11-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    schedule_2 = CharacteristicBytes(
        "98bd0c12-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    schedule_3 = CharacteristicBytes(
        "98bd0c13-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    schedule_4 = CharacteristicBytes(
        "98bd0c14-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    schedule_5 = CharacteristicBytes(
        "98bd0c15-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    schedule_6 = CharacteristicBytes(
        "98bd0c16-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    schedule_7 = CharacteristicBytes(
        "98bd0c17-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    schedule_8 = CharacteristicBytes(
        "98bd0c18-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    schedule_9 = CharacteristicBytes(
        "98bd0c19-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    schedule_10 = CharacteristicBytes(
        "98bd0c1a-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    schedule_11 = CharacteristicBytes(
        "98bd0c1b-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    schedule_12 = CharacteristicBytes(
        "98bd0c1c-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    schedule_13 = CharacteristicBytes(
        "98bd0c1d-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    schedule_14 = CharacteristicBytes(
        "98bd0c1e-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    schedule_15 = CharacteristicBytes(
        "98bd0c1f-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )


class Schedule(Service, ABC):
    products = set(ProductType) - {ProductType.AQUA_CONTOURS}

    def __init_subclass__(cls, *, instance: int, **kwargs):
        def _uuid(offset: int) -> str:
            return f"98bd0c{0x10 * instance + offset:02x}-0b0e-421a-84e5-ddbf75dc6de4"

        cls.uuid = _uuid(0)
        cls.start_time = CharacteristicLong(_uuid(1), name="Start Time")
        cls.duration = CharacteristicLong(_uuid(2), name="Duration")
        cls.weekdays = CharacteristicWeekday(_uuid(3), name="Weekdays")
        cls.valve_link = CharacteristicBytes(_uuid(4), name="Valve Link")
        cls.active = CharacteristicBool(_uuid(5), name="Active")
        cls.sensor_link = CharacteristicBool(_uuid(6), name="Sensor Link")
        super().__init_subclass__(**kwargs)


class Schedule_1(Schedule, instance=1):
    pass


class Schedule_2(Schedule, instance=2):
    pass


class Schedule_3(Schedule, instance=3):
    pass


class Schedule_4(Schedule, instance=4):
    pass


class Schedule_5(Schedule, instance=5):
    pass


class DeviceInformation(Service):
    uuid = "0000180a-0000-1000-8000-00805f9b34fb"
    model_number = CharacteristicString("00002a24-0000-1000-8000-00805f9b34fb")
    firmware_version = CharacteristicString("00002a26-0000-1000-8000-00805f9b34fb")
    manufacturer_name = CharacteristicString("00002a29-0000-1000-8000-00805f9b34fb")


class Battery(Service):
    uuid = "98bd180f-0b0e-421a-84e5-ddbf75dc6de4"

    battery_level = CharacteristicInt("98bd2a19-0b0e-421a-84e5-ddbf75dc6de4")


class Sensor(Service):
    uuid = "98bd0010-0b0e-421a-84e5-ddbf75dc6de4"

    value = CharacteristicInt("98bd0011-0b0e-421a-84e5-ddbf75dc6de4")
    connected_state = CharacteristicBool("98bd0012-0b0e-421a-84e5-ddbf75dc6de4")
    type = CharacteristicString("98bd0013-0b0e-421a-84e5-ddbf75dc6de4")
    threshold = CharacteristicInt("98bd0014-0b0e-421a-84e5-ddbf75dc6de4")
    battery_level = CharacteristicInt("98bd0015-0b0e-421a-84e5-ddbf75dc6de4")
    measurement_timestamp = CharacteristicTime("98bd0016-0b0e-421a-84e5-ddbf75dc6de4")
    force_measurement = CharacteristicInt("98bd0017-0b0e-421a-84e5-ddbf75dc6de4")


class WateringHistory(Service):
    uuid = "98bd0d10-0b0e-421a-84e5-ddbf75dc6de4"
    products = set(ProductType) - {ProductType.AQUA_CONTOURS}

    timestamp_array = CharacteristicTimeArray("98bd0d11-0b0e-421a-84e5-ddbf75dc6de4")
    timestamp_count = CharacteristicInt("98bd0d12-0b0e-421a-84e5-ddbf75dc6de4")
    skip_reason = CharacteristicBytes("98bd0d13-0b0e-421a-84e5-ddbf75dc6de4")
    watering_duration = CharacteristicLongArray("98bd0d14-0b0e-421a-84e5-ddbf75dc6de4")
    watering_skipped = CharacteristicBool("98bd0d15-0b0e-421a-84e5-ddbf75dc6de4")
    skipped_schedule_number = CharacteristicInt("98bd0d16-0b0e-421a-84e5-ddbf75dc6de4")
    water_control_error = CharacteristicInt("98bd0d17-0b0e-421a-84e5-ddbf75dc6de4")
    watering_pause = CharacteristicLong("98bd0d18-0b0e-421a-84e5-ddbf75dc6de4")
    seasonal_adjust = CharacteristicInt("98bd0d19-0b0e-421a-84e5-ddbf75dc6de4")
    rain_sensitivity = CharacteristicInt("98bd0d1a-0b0e-421a-84e5-ddbf75dc6de4")


class ErrorHistory(Service):
    uuid = "98bdeeee-0b0e-421a-84e5-ddbf75dc6de4"

    error_id = CharacteristicBytes("98bdeeef-0b0e-421a-84e5-ddbf75dc6de4")
    error_count = CharacteristicInt("98bdeef0-0b0e-421a-84e5-ddbf75dc6de4")


class Pump(Service):
    uuid = "98bd0100-0b0e-421a-84e5-ddbf75dc6de4"

    status = CharacteristicInt("98bd0101-0b0e-421a-84e5-ddbf75dc6de4")
    tank_preassure = CharacteristicUInt16("98bd0102-0b0e-421a-84e5-ddbf75dc6de4")
    flow_rate = CharacteristicUInt16("98bd0103-0b0e-421a-84e5-ddbf75dc6de4")
    ptu_mode = CharacteristicInt("98bd0104-0b0e-421a-84e5-ddbf75dc6de4")
    leakage_detection = CharacteristicBool("98bd0105-0b0e-421a-84e5-ddbf75dc6de4")
    min_preassure = CharacteristicUInt16("98bd0106-0b0e-421a-84e5-ddbf75dc6de4")
    max_preassure = CharacteristicUInt16("98bd0107-0b0e-421a-84e5-ddbf75dc6de4")
    child_lock = CharacteristicBool("98bd0108-0b0e-421a-84e5-ddbf75dc6de4")
    filter_reminder = CharacteristicInt("98bd0109-0b0e-421a-84e5-ddbf75dc6de4")
    direct_start = CharacteristicBool("98bd010a-0b0e-421a-84e5-ddbf75dc6de4")
    max_runtime = CharacteristicInt("98bd010b-0b0e-421a-84e5-ddbf75dc6de4")
    safety_pump_time = CharacteristicInt("98bd010c-0b0e-421a-84e5-ddbf75dc6de4")
    cool_down_timer = CharacteristicUInt16("98bd010d-0b0e-421a-84e5-ddbf75dc6de4")
    water_temperature = CharacteristicInt("98bd010e-0b0e-421a-84e5-ddbf75dc6de4")
    error_code = CharacteristicBytes("98bd010f-0b0e-421a-84e5-ddbf75dc6de4")
    user_motor_runtime = CharacteristicLong("98bd0110-0b0e-421a-84e5-ddbf75dc6de4")
    total_motor_runtime = CharacteristicLong("98bd0111-0b0e-421a-84e5-ddbf75dc6de4")


class Spray(Service):
    uuid = "98bd0110-0b0e-421a-84e5-ddbf75dc6de4"
    variant = "1"

    distance = CharacteristicUInt16("98bd0111-0b0e-421a-84e5-ddbf75dc6de4", variant="1")
    sector = CharacteristicUInt16("98bd0112-0b0e-421a-84e5-ddbf75dc6de4", variant="1")
    current_distance = CharacteristicUInt16(
        "98bd0113-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    current_sector = CharacteristicUInt16(
        "98bd0114-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    watering_mode_error = CharacteristicInt(
        "98bd0115-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )


class AquaContourErrorCode(IntEnum):
    NO_ERROR = 0
    NO_WATER = 1
    NOT_ENOUGH_WATER = 2
    CHARGER_ERROR = 3
    SPRINKLER_MOTOR_ERROR = 4
    VALVE_MOTOR_ERROR = 5
    ROTATION_SENSOR_ERROR = 6
    FLASH_ERROR = 7


class EventHistory(Service):
    uuid = "98bd0120-0b0e-421a-84e5-ddbf75dc6de4"
    history = CharacteristicBytes("98bd0121-0b0e-421a-84e5-ddbf75dc6de4")
    error = CharacteristicErrorData(
        "98bd0122-0b0e-421a-84e5-ddbf75dc6de4", enum=AquaContourErrorCode
    )


class AquaContourOperationMode(IntEnum):
    ACTIVE = 1
    MANUAL_MODE = 3
    PRE_WINTER = 4
    DEEP_SLEEP = 6


class AquaContour(Service):
    uuid = "98bd0a10-0b0e-421a-84e5-ddbf75dc6de4"
    products = {ProductType.AQUA_CONTOURS}

    unix_timestamp = CharacteristicTime("98bd0a11-0b0e-421a-84e5-ddbf75dc6de4")
    custom_device_name = CharacteristicNullStringUf8(
        "98bd0a12-0b0e-421a-84e5-ddbf75dc6de4"
    )
    frost_warning = CharacteristicBool("98bd0a15-0b0e-421a-84e5-ddbf75dc6de4")
    active_contour = CharacteristicBytes("98bd0a16-0b0e-421a-84e5-ddbf75dc6de4")
    operation_mode = CharacteristicIntEnum(
        "98bd0a17-0b0e-421a-84e5-ddbf75dc6de4", enum=AquaContourOperationMode
    )
    factory_reset = CharacteristicInt("98bd0a18-0b0e-421a-84e5-ddbf75dc6de4")


class AquaContourWateringMode(IntEnum):
    PREVIEW = -3
    SETUP_MODE = -1
    REST = 0
    CONTOUR_1 = 1
    CONTOUR_2 = 2
    CONTOUR_3 = 3
    CONTOUR_4 = 4
    CONTOUR_5 = 5


class AquaContourWatering(Service):
    uuid = "98bd0d10-0b0e-421a-84e5-ddbf75dc6de4"
    variant = "1"
    products = {ProductType.AQUA_CONTOURS}
    watering_active = CharacteristicIntEnum(
        "98bd0d11-0b0e-421a-84e5-ddbf75dc6de4",
        variant="1",
        enum=AquaContourWateringMode,
    )
    remaining_watering_time = CharacteristicLong(
        "98bd0d12-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    manual_watering_time = CharacteristicLong(
        "98bd0d13-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    activation_reason = CharacteristicInt(
        "98bd0d14-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    watering_skipped = CharacteristicBool(
        "98bd0d15-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    skipped_schedule_number = CharacteristicInt(
        "98bd0d15-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    watering_control_error = CharacteristicInt(
        "98bd0d16-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    skipped_reason = CharacteristicInt(
        "98bd0d17-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    watering_pause = CharacteristicInt(
        "98bd0d18-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    seasonal_adjust = CharacteristicInt(
        "98bd0d19-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )
    rain_sensitivity = CharacteristicInt(
        "98bd0d1a-0b0e-421a-84e5-ddbf75dc6de4", variant="1"
    )


class FlowStatistics(Service):
    uuid = "98bd0e10-0b0e-421a-84e5-ddbf75dc6de4"
    overall = CharacteristicLong("98bd0e16-0b0e-421a-84e5-ddbf75dc6de4")
    resettable = CharacteristicLong("98bd0e17-0b0e-421a-84e5-ddbf75dc6de4")
    last_reset = CharacteristicTime("98bd0e18-0b0e-421a-84e5-ddbf75dc6de4")
    current = CharacteristicUInt16("98bd0e19-0b0e-421a-84e5-ddbf75dc6de4")


class AquaContourPosition(Service):
    uuid = "98bd0130-0b0e-421a-84e5-ddbf75dc6de4"
    products = {ProductType.AQUA_CONTOURS}

    active_position = CharacteristicInt("98bd0132-0b0e-421a-84e5-ddbf75dc6de4")
    position_contour_mask = CharacteristicBytes("98bd0135-0b0e-421a-84e5-ddbf75dc6de4")
    position_name_1 = CharacteristicNullStringUf8(
        "98bd013a-0b0e-421a-84e5-ddbf75dc6de4"
    )
    position_name_2 = CharacteristicNullStringUf8(
        "98bd013b-0b0e-421a-84e5-ddbf75dc6de4"
    )
    position_name_3 = CharacteristicNullStringUf8(
        "98bd013c-0b0e-421a-84e5-ddbf75dc6de4"
    )
    position_name_4 = CharacteristicNullStringUf8(
        "98bd013d-0b0e-421a-84e5-ddbf75dc6de4"
    )
    position_name_5 = CharacteristicNullStringUf8(
        "98bd013e-0b0e-421a-84e5-ddbf75dc6de4"
    )


class AquaContourBattery(Service):
    uuid = "0000180f-0000-1000-8000-00805f9b34fb"
    products = {ProductType.AQUA_CONTOURS}

    battery_level = CharacteristicInt("00002a19-0000-1000-8000-00805f9b34fb")
    battery_level_status = CharacteristicInt("00002bed-0000-1000-8000-00805f9b34fb")


class Reset(Service):
    uuid = "98bdff00-0b0e-421a-84e5-ddbf75dc6de4"

    factory_reset = CharacteristicBool("98bdff01-0b0e-421a-84e5-ddbf75dc6de4")


class Oad(Service):
    uuid = "f000ffd0-0451-4000-b000-000000000000"

    enable_oad = CharacteristicBool("f000ffd1-0451-4000-b000-000000000000")


class Fota(Service):
    uuid = "f000ffc0-0451-4000-b000-000000000000"

    image_identify = CharacteristicBytes("f000ffc1-0451-4000-b000-000000000000")
    image_block_id = CharacteristicBytes("f000ffc2-0451-4000-b000-000000000000")
    control_point = CharacteristicBytes("f000ffc5-0451-4000-b000-000000000000")



from datetime import datetime

from gardena_bluetooth.parse import (
    ManufacturerData,
    ProductGroup,
    ProductType,
    CharacteristicString,
    CharacteristicNullStringUf8,
    CharacteristicNullString,
    CharacteristicIntEnum,
    CharacteristicErrorData,
)
from enum import IntEnum


def test_manufacturer_data():
    raw = b"\x02\x07d\x02\x05\x01\x02\x08\x00\x02" b"\t\x01\x04\x06\x12\x00\x01"
    data = ManufacturerData.decode(raw)
    assert data == ManufacturerData(
        pairable=True, serial=None, group=ProductGroup.WATER_CONTROL, model=0, variant=1
    )

    assert ProductType.from_manufacturer_data(data) == ProductType.WATER_COMPUTER

    raw = (
        b"\x02\x07d\x02\x05\x01\x02\x08\x00\x02"
        b"\t\x01\x04\x06\x20\x00\x01\x05\x04\x01\x02\x03\x04"
    )
    data = ManufacturerData.decode(raw)
    assert data == ManufacturerData(
        pairable=True, serial=0x04030201, group=32, model=0, variant=1
    )

    assert ProductType.from_manufacturer_data(data) is ProductType.UNKNOWN

    raw = (
        b"\x05\x04\x14\x18\x00\x00\x02\x05\x01\x04\x06\x11\x02\x02"
        b"\t\x01\x04\x06\x20\x00\x01\x05\x04\x01\x02\x03\x04"
    )
    data = ManufacturerData.decode(raw)
    assert data == ManufacturerData(
        pairable=True, serial=6164, group=17, model=2, variant=2
    )

    assert ProductType.from_manufacturer_data(data) is ProductType.PRESSURE_TANKS


def test_parse_manufacturer_data_segmented():
    """Test segmented manufacturer data."""

    data = ManufacturerData()
    raw = bytes.fromhex("0205000406121001")
    data.update(raw)

    assert data == ManufacturerData(
        pairable=False,
        group=ProductGroup.WATER_CONTROL,
        model=16,
        variant=1,
    )

    raw = bytes.fromhex("0504f162c103")
    data.update(raw)

    assert data == ManufacturerData(
        pairable=False,
        serial=63005425,
        group=ProductGroup.WATER_CONTROL,
        model=16,
        variant=1,
    )
    assert ProductType.from_manufacturer_data(data) is ProductType.AQUA_CONTOURS


def test_string_firmware_invalid():
    raw = b"abc\xe4"
    data = CharacteristicString.decode(raw)
    assert data == "abc�"


def test_string_nulled_utf8():
    raw = "åäö".encode("utf-8") + b"\x00junk"
    data = CharacteristicNullStringUf8.decode(raw)
    assert data == "åäö"


def test_string_nulled():
    raw = b"abc\x00junk"
    data = CharacteristicNullString.decode(raw)
    assert data == "abc"


def test_enum():
    class Values(IntEnum):
        A = 0
        B = 2

    char = CharacteristicIntEnum("", enum=Values)
    raw = b"\x00"
    data = char.decode(raw)
    assert data is Values.A
    raw = b"\x01"
    data = char.decode(raw)
    assert data == 1
    assert char.encode(Values.B) == b"\x02"


def test_error_code():
    class Errors(IntEnum):
        A = 0
        B = 1
        C = 2

    char = CharacteristicErrorData("", enum=Errors)
    data = char.decode(b"\x01\x01\xc3,\xafi\x01")
    assert data.error_code is Errors.B
    assert data.time_stamp == datetime(2026, 3, 9, 20, 25, 39)
    assert data.current_event_index == 1
    assert data.total_events == 1



from __future__ import annotations

from bleak.uuids import register_uuids

from .const import ScanService
from .parse import Service, Characteristic


def prefix(name: str) -> str:
    return f"Gardena {name}"


register_uuids(
    {
        uuid: prefix(", ".join(service.__name__ for service in services))
        for uuid, services in Service.registry.items()
    }
)

register_uuids(
    {
        uuid: prefix(", ".join(char.name for char in chars))
        for uuid, chars in Characteristic.registry.items()
    }
)

register_uuids({ScanService: "Husqvarna"})
